 /*
  * Mike's Raycaster 0.5 - A demonstration of raycasting in java
  * 	http://mike.bitrevision.com/raycaster/
  * 
  * Copyright (c) 2008-2012 Michael Billington <michael.billington@gmail.com>
  * 
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
  * in the Software without restriction, including without limitation the rights
  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  * copies of the Software, and to permit persons to whom the Software is
  * furnished to do so, subject to the following conditions:

  * The above copyright notice and this permission notice shall be included in
  * all copies or substantial portions of the Software.

  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  * THE SOFTWARE. */

import java.awt.*;

class Engine {
	/* Class that contains that contains rendering algorithms (hit space for 2D mode for debugging) */
	static boolean DISPLAY_2D = false;
	static Player player;
	static PlayerPos pos;
	static GameMap map;
	static UserInput input;

	public static void initialise(String mapPath) {
		player = new Player();
		map = new GameMap(mapPath);
	}
	
	public static void movePlayer() {
		/* Move player according to keyboard state */
		if(input.getKeyState(input.KEY_UP)) {
			player.move(5);
		} else if(input.getKeyState(input.KEY_DOWN)) {
			player.move(-5);
		}
		
		if(input.getKeyState(input.KEY_LEFT)) {
			player.turn(-Math.PI / 32);
		} else if(input.getKeyState(input.KEY_RIGHT)) {
			player.turn(Math.PI / 32);
		}
	}

	public static void renderFrame(Graphics board, Dimension d) {
		if(DISPLAY_2D) {
			renderFrame_2D(board, d);
		} else {
			renderFrame_3D(board, d);
		}
	}


	
	public static void renderFrame_3D(Graphics board, Dimension d) {
		int i, lineHeight;
		double angleOffset;
		RayResult ray = new RayResult();
		pos = player.getPos();
		
		/* Prepare board */
		board.setColor(Color.black);
		board.fillRect(0, 0, d.width, d.height);

		for(i = 0; i < d.width; i++) {
			angleOffset = pos.angle + ((double) i / d.width * (Math.PI / 3)) -  Math.PI / 6;
			angleOffset = Player.normaliseAngle(angleOffset);
			ray = castRay(pos, angleOffset, ray);
			lineHeight = (int)((double) d.height * (70 / ray.length));
			board.setColor(getBlockColour(ray.length, ray.hitTile));
			board.drawLine(i, (d.height - lineHeight) / 2, i, (d.height - lineHeight) / 2 + lineHeight);
		}
	}
	
	static Color getBlockColour(double rayLength, byte hitTile) {
		int darkness = (int) rayLength / 3;
		if(darkness > 255) { darkness = 255; }
		darkness = 255 - darkness;
		switch(hitTile) {
			case '1':
				return new Color(darkness, 0, 0);
			case '2':
				return new Color(0, darkness, 0);
			case '3':
				return new Color(0, 0, darkness);
			case '4':
				return new Color(darkness, 0, darkness);
			case '5':
				return new Color(darkness, darkness, 0);
			case '6':
				return new Color(0, darkness, darkness);
			case '7':
				return new Color(darkness / 2, darkness / 2, darkness / 2);
			case 0:
			default:
				return new Color(darkness, darkness, darkness);
		}
	}

	static RayResult castRay(PlayerPos source, double angleOffset, RayResult progress) {
		/* Actual ray-casting routine */
		progress.length = 0;
		
		int curPosX = source.PosX, dPosX;
		int curPosY = source.PosY, dPosY;
		int curBlockX = source.BlockX, dBlockX = 0;
		int curBlockY = source.BlockY, dBlockY = 0;
		double compareAngle;
		
		double sine = Math.sin(angleOffset), cosine = Math.cos(angleOffset), diagonal;
		/* We split into four quadrants, each with three possible outcomes (diagonal or two sides)
		 * 	They are related by symmetry, but it is easy enough just to enumerate them.
		 */
		do {
			if(sine < 0) {
				if(cosine >= 0) {
					/* First quadrant */
					diagonal = angleTo(curPosX, curPosY, Raycaster.BLOCK_SIZE, -1);
					if(angleOffset == diagonal) {
						/* Hit top-right diagonal */
						dPosX = 0;
						dPosY = 0;
					} else if (angleOffset > diagonal) {
						/* Hit right wall (first quadrant) */
						dPosX = Raycaster.BLOCK_SIZE - curPosX;
						dPosY = (int)((double)dPosX * Math.tan(angleOffset));
						dBlockY = 0;
						dBlockX = 1;
					} else {
						/* Hit top wall (first quadrant) */
						dPosY = -curPosY;
						dPosX = (int)((double)dPosY * Math.tan(Math.PI / 2 - angleOffset));
						dBlockY = -1;
						dBlockX = 0;
					}
				} else {
					/* Second Quadrant */
					diagonal = angleTo(curPosX, curPosY, -1, -1);
					compareAngle = angleOffset - Math.PI;
					if(compareAngle == diagonal) {
						/* Hit top-right diagonal */
						dPosX = 0;
						dPosY = 0;
					} else if (compareAngle < diagonal) {
						/* Hit left wall (second quadrant) */
						dPosX = -curPosX;
						dPosY = (int)((double)dPosX * Math.tan(compareAngle));
						dBlockY = 0;
						dBlockX = -1;
					} else {
						/* Hit top wall (second quadrant) */
						dPosY = -curPosY;
						dPosX = (int)((double)dPosY * Math.tan(Math.PI / 2 - compareAngle));
						dBlockY = -1;
						dBlockX = 0;
					}
				}
			} else {
				if(cosine <= 0) {
					/* Third quadrant */
					diagonal = angleTo(curPosX, curPosY, -1, Raycaster.BLOCK_SIZE) - Math.PI;
					compareAngle = angleOffset;
					if(compareAngle == diagonal) {
						/* Hit top-right diagonal */
						dPosX = 0;
						dPosY = 0;
					} else if (compareAngle > diagonal) {
						/* Hit left wall (third quadrant) */
						dPosX = -curPosX;
						dPosY = (int)((double)dPosX * Math.tan(compareAngle));
						dBlockY = 0;
						dBlockX = -1;
					} else {
						/* Hit bottom wall (third quadrant) */
						dPosY = Raycaster.BLOCK_SIZE - curPosY;
						dPosX = (int)((double)dPosY * Math.tan(Math.PI / 2 - compareAngle));
						dBlockY = 1;
						dBlockX = 0;
					}
				} else {
					/* Fourth quadrant */
					diagonal = angleTo(curPosX, curPosY, Raycaster.BLOCK_SIZE, Raycaster.BLOCK_SIZE);
					compareAngle = angleOffset;
	
					if(compareAngle == diagonal) {
						/* Hit top-right diagonal */
						dPosX = 0;
						dPosY = 0;
					} else if (compareAngle < diagonal) {
						/* Hit right wall (fourth quadrant) */
						dPosX = Raycaster.BLOCK_SIZE - curPosX;
						dPosY = (int)((double)dPosX * Math.tan(compareAngle));
						dBlockY = 0;
						dBlockX = 1;
					} else {
						/* Hit bottom wall (fourth quadrant) */
						dPosY = Raycaster.BLOCK_SIZE - curPosY;
						dPosX = (int)((double)dPosY * Math.tan(Math.PI / 2 - compareAngle));
						dBlockY = 1;
						dBlockX = 0;
					}
				}
			}
			
			progress.length += Math.sqrt(Math.pow(dPosX, 2) + Math.pow(dPosY, 2));
			curBlockX += dBlockX;
			curBlockY += dBlockY;
			curPosX = (curPosX + dPosX + Raycaster.BLOCK_SIZE) % Raycaster.BLOCK_SIZE;
			curPosY = (curPosY + dPosY + Raycaster.BLOCK_SIZE) % Raycaster.BLOCK_SIZE;
			if(dBlockX < 0) {
				curPosX = Raycaster.BLOCK_SIZE - 1;
			}
			if(dBlockY < 0) {
				curPosY = Raycaster.BLOCK_SIZE - 1;
			}
		} while(curBlockX >= 0 && curBlockY >= 0 && curBlockX < map.width && curBlockY < map.height && map.tile[curBlockX][curBlockY] == ' ');
		
		if(curBlockX >= 0 && curBlockY >= 0 && curBlockX < map.width && curBlockY < map.height) {
			progress.hitTile = map.tile[curBlockX][curBlockY];
		} else {	
			progress.hitTile = '0';
		}

		return progress;
	}

	static double angleTo(int x1, int y1, int x2, int y2) {
		/* Uses tan-1 to find the angle between two points */
		double angle;
		if(x2 - x1 == 0) {
			/* Need to avoid division by zero of course.. */
			if(y2 >= y2) {
				angle = Math.PI / 2;
			} else {
				angle = -Math.PI / 2;
			}
		} else {
			double m = (double) (y2 - y1) / (x2 - x1);
			angle = Math.atan(m);
		}
		return Player.normaliseAngle(angle);
	}
	
	public static void renderFrame_2D(Graphics board, Dimension d) {
		/* 2D birds-eye view rendering */
		int x, y, i;
		int x2, y2;
		double angleOffset;
		pos = player.getPos();

		board.setColor(Color.blue);
		board.fillRect(0, 0, d.width, d.height);
	
		/* Draw grid over the board */
		board.setColor(Color.green);
		for(x = Raycaster.BLOCK_SIZE; x < d.width; x += Raycaster.BLOCK_SIZE) {
			board.drawLine(x, 0, x, d.height - 1);
		}
		for(y = Raycaster.BLOCK_SIZE; y < d.height; y += Raycaster.BLOCK_SIZE) {
			board.drawLine(0, y, d.width - 1, y);
		}
		
		board.setColor(Color.red);
		for(y = 0; y < map.height; y++) {
			for(x = 0; x < map.width; x++) {
				if(map.tile[x][y] != ' ') {
					/* cross out bad squares */
					board.drawLine(x * Raycaster.BLOCK_SIZE, y * Raycaster.BLOCK_SIZE, (x + 1) * Raycaster.BLOCK_SIZE, (y + 1) * Raycaster.BLOCK_SIZE);
					board.drawLine(x * Raycaster.BLOCK_SIZE, (y + 1) * Raycaster.BLOCK_SIZE, (x + 1) * Raycaster.BLOCK_SIZE, y * Raycaster.BLOCK_SIZE);
				}
			}
		}
		
		x = pos.PosX + pos.BlockX * Raycaster.BLOCK_SIZE;
		y = pos.PosY + pos.BlockY * Raycaster.BLOCK_SIZE;
		board.fillOval(x - 5, y - 5, 10, 10);
		
		board.setColor(Color.yellow);
		RayResult ray = new RayResult();
		for(i = 0; i < d.width; i++) {
			/* Cast each ray. */
			angleOffset = pos.angle + ((double) i / d.width * (Math.PI / 3)) -  Math.PI / 6;
			angleOffset = Player.normaliseAngle(angleOffset);
			ray = castRay(pos, angleOffset, ray);
			x2 = x + (int)(Math.cos(angleOffset) * ray.length);
			y2 = y + (int)(Math.sin(angleOffset) * ray.length);
			board.drawLine(x, y, x2, y2);
		}
	}
}
